# frozen_string_literal: true

require "interjectable/version"

module Interjectable
  MethodAlreadyDefined = Class.new(StandardError)

  def self.included(mod)
    mod.send(:extend, ClassMethods)
  end

  def self.extended(mod)
    mod.send(:extend, ClassMethods)
  end

  module ClassMethods
    # Defines a helper methods on instances that memoize values per-instance.
    #
    # Calling a second time is an error. Use `#test_inject` for overriding in
    # RSpec tests. You need to `require "interjectable/rspec"` to use
    # `#test_inject`. See the README.md.
    #
    # Similar to writing
    #
    #   attr_writer :dependency
    #
    #   def dependency
    #     @dependency ||= instance_eval(&default_block)
    #   end
    def inject(dependency, &default_block)
      if instance_methods(false).include?(dependency)
        raise MethodAlreadyDefined, "#{dependency} is already defined"
      end

      attr_writer dependency

      define_method(dependency) do
        ivar_name = :"@#{dependency}"
        if instance_variable_defined?(ivar_name)
          instance_variable_get(ivar_name)
        else
          instance_variable_set(ivar_name, instance_eval(&default_block))
        end
      end
    end

    # Defines helper methods on instances that memoize values per-class.
    # (shared across all instances of a class, including instances of
    # subclasses).
    #
    # Calling a second time is an error. Use `#test_inject` for overriding in
    # RSpec tests. You need to `require "interjectable/rspec"` to use
    # `#test_inject`. See the README.md.
    #
    # Similar to writing
    #
    #   cattr_writer :dependency
    #
    #   def dependency
    #     @@dependency ||= instance_eval(&default_block)
    #   end
    def inject_static(dependency, &default_block)
      if instance_methods(false).include?(dependency) || methods(false).include?(dependency)
        raise MethodAlreadyDefined, "#{dependency} is already defined"
      end

      cvar_name = :"@@#{dependency}"
      setter = :"#{dependency}="

      define_method(setter) do |value|
        self.class.send(setter, value)
      end

      define_singleton_method(setter) do |value|
        class_variable_set(cvar_name, value)
      end

      define_method(dependency) do
        self.class.send(dependency)
      end

      define_singleton_method(dependency) do
        if class_variable_defined?(cvar_name)
          class_variable_get(cvar_name)
        else
          class_variable_set(cvar_name, instance_eval(&default_block))
        end
      end
    end
  end
end
